/* 
   Unix SMB/Netbios implementation.
   Version 2.0.
   LDAP builtin group database for SAMBA
   Copyright (C) Matthew Chapman 1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
   
*/

#include "includes.h"
#include "libsamba.h"
#include "samdb.h"
#include "others.h"

#ifdef USE_PASSDB_LDAP

#include <lber.h>
#include <ldap.h>


/* Internal state */
extern LDAP *ldap_struct;
extern LDAPMessage *ldap_results;
extern LDAPMessage *ldap_entry;

/* Static structure filled for requests */
static LOCAL_GRP localgrp;


/***************************************************************
  Get group and membership information.
 ****************************************************************/

static LOCAL_GRP *ldapbuiltin_getgrp(LOCAL_GRP *group,
		     LOCAL_GRP_MEMBER **members, int *num_membs)
{
	fstring temp;
	char **values;
	LOCAL_GRP_MEMBER *memblist;
	char *value, *sep;
	int i;

	if(!ldap_entry)
		return NULL;

	if(!ldap_get_attribute("cn", group->name)) {
		DEBUG(0, ("Missing cn\n"));
                return NULL; }
	
	DEBUG(2,("Retrieving builtin alias [%s]\n", group->name));

        if(ldap_get_attribute("rid", temp)) {
		group->rid = strtol(temp, NULL, 16);
	} else {
		DEBUG(0, ("Missing rid\n"));
		return NULL;
	}

	if(!ldap_get_attribute("description", group->comment))
		group->comment[0] = 0;

	if(!members || !num_membs) {
		ldap_entry = ldap_next_entry(ldap_struct, ldap_entry);
		return group;
	}

#ifdef ENABLE_OLD_LDAP_SCHEMA
	if((values = ldap_get_values(ldap_struct, ldap_entry, "member"))) {
#else
	if((values = ldap_get_values(ldap_struct, ldap_entry, "sambaMember"))) {
#endif /* ENABLE_OLD_LDAP_SCHEMA */

		*num_membs = i = ldap_count_values(values);
		*members = memblist = malloc(i * sizeof(LOCAL_GRP_MEMBER));

		do {
			value = values[--i];

		        if(!(sep = strchr(value, ','))) {
				DEBUG(0, ("Malformed alias member\n"));
				return NULL;
			}
			*(sep++) = 0;
			fstrcpy(memblist[i].name, value);

			if(!(value = strchr(sep, ','))) {
				DEBUG(0, ("Malformed alias member\n"));
				return NULL;
			}
			*(value++) = 0;
			string_to_sid(&memblist[i].sid, sep);

		} while(i > 0);

		ldap_value_free(values);

	} else {
		*num_membs = 0;
		*members = NULL;
	}

	return group;
}


/************************************************************************
  Queues the necessary modifications to save a LOCAL_GRP structure
 ************************************************************************/

static void ldapbuiltin_grpmods(LOCAL_GRP *group, LDAPMod ***mods,
			int operation)
{
	fstring temp;

	*mods = NULL;

	if(operation == LDAP_MOD_ADD) { /* immutable attributes */
		ldap_make_mod(mods, LDAP_MOD_ADD, "objectClass", "sambaBuiltin");
		ldap_make_mod(mods, LDAP_MOD_ADD, "cn", group->name);

		slprintf(temp, sizeof(temp)-1, "%x", group->rid);
		ldap_make_mod(mods, LDAP_MOD_ADD, "rid", temp);
	}

	ldap_make_mod(mods, operation, "description", group->comment);
}


/************************************************************************
  Create a builtin alias member entry
 ************************************************************************/

static BOOL ldapbuiltin_memmods(const DOM_SID *user_sid, LDAPMod ***mods,
			      int operation)
{
	pstring member;
	pstring sid_str;
	fstring name;
	uint32 type;

	if (lookup_sid(user_sid, name, &type))
		return (False);
	sid_to_string(sid_str, user_sid);

	slprintf(member, sizeof(member)-1, "%s,%s,%d", name, sid_str, type);

	*mods = NULL;
#ifdef ENABLE_OLD_LDAP_SCHEMA
	ldap_make_mod(mods, operation, "member", member);
#else
	ldap_make_mod(mods, operation, "sambaMember", member);
#endif /* ENABLE_OLD_LDAP_SCHEMA */
	return True;
}


/***************************************************************
  Begin/end smbgrp enumeration.
 ****************************************************************/

static void *ldapbuiltin_enumfirst(BOOL update)
{
	if (!ldap_connect())
		return NULL;

	ldap_search_for("objectClass=sambaBuiltin");

	return ldap_struct;
}

static void ldapbuiltin_enumclose(void *vp)
{
	ldap_disconnect();
}


/*************************************************************************
  Return limited smb_passwd information, and group membership.
 *************************************************************************/

static LOCAL_GRP *ldapbuiltin_getgrpbynam(const char *name,
	       LOCAL_GRP_MEMBER **members, int *num_membs)
{
	fstring filter;
	LOCAL_GRP *ret;

	if(!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(cn=%s)(objectClass=sambaBuiltin))", name);
	ldap_search_for(filter);

	ret = ldapbuiltin_getgrp(&localgrp, members, num_membs);

	ldap_disconnect();
	return ret;
}

static LOCAL_GRP *ldapbuiltin_getgrpbyrid(uint32 grp_rid,
	       LOCAL_GRP_MEMBER **members, int *num_membs)
{
	fstring filter;
	LOCAL_GRP *ret;

	if(!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(rid=%x)(objectClass=sambaBuiltin))", grp_rid);
	ldap_search_for(filter);
	ret = ldapbuiltin_getgrp(&localgrp, members, num_membs);

	ldap_disconnect();
	return ret;
}

static LOCAL_GRP *ldapbuiltin_getcurrentgrp(void *vp,
	       LOCAL_GRP_MEMBER **members, int *num_membs)
{
	return ldapbuiltin_getgrp(&localgrp, members, num_membs);
}


/*************************************************************************
  Add/modify/delete builtin aliases.
 *************************************************************************/

static BOOL ldapbuiltin_addgrp(LOCAL_GRP *group)
{
	LDAPMod **mods;

	if (!ldap_allocaterid(&group->rid))
	{
	    DEBUG(0,("RID generation failed\n"));
	    return (False);
	}

	ldapbuiltin_grpmods(group, &mods, LDAP_MOD_ADD); 
	return ldap_commit_mods("cn", group->name, mods, True,
				lp_ldap_builtin_subcontext())
		&& ldap_incr_seq_num();
}

static BOOL ldapbuiltin_modgrp(LOCAL_GRP *group)
{
	LDAPMod **mods;

	ldapbuiltin_grpmods(group, &mods, LDAP_MOD_REPLACE);
	return ldap_commit_mods("cn", group->name, mods, False,
				lp_ldap_builtin_subcontext())
		&& ldap_incr_seq_num();
}

static BOOL ldapbuiltin_delgrp(uint32 grp_rid)
{
	fstring filter;
	char *dn;
	int err;

	if (!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(rid=%x)(objectClass=sambaBuiltin))", grp_rid);
	ldap_search_for(filter);

	if (!ldap_entry || !(dn = ldap_get_dn(ldap_struct, ldap_entry)))
	{
		ldap_disconnect();
		return (False);
	}

	err = ldap_delete_s(ldap_struct, dn);
	free(dn);
	ldap_disconnect();

	if (err != LDAP_SUCCESS)
	{
		DEBUG(0, ("delete: %s\n", ldap_err2string(err)));
		return (False);
	}

	return True;
}


/*************************************************************************
  Add users to/remove users from aliases.
 *************************************************************************/

static BOOL ldapbuiltin_addmem(uint32 grp_rid, const DOM_SID *user_sid)
{
	LDAPMod **mods;
        fstring rid_str;

	slprintf(rid_str, sizeof(rid_str)-1, "%x", grp_rid);

	if(!ldapbuiltin_memmods(user_sid, &mods, LDAP_MOD_ADD))
		return (False);

	return ldap_commit_mods("rid", rid_str, mods, False,
				lp_ldap_builtin_subcontext())
		&& ldap_incr_seq_num();
}

static BOOL ldapbuiltin_delmem(uint32 grp_rid, const DOM_SID *user_sid)
{
	LDAPMod **mods;
        fstring rid_str;

	slprintf(rid_str, sizeof(rid_str)-1, "%x", grp_rid);

	if(!ldapbuiltin_memmods(user_sid, &mods, LDAP_MOD_DELETE))
		return (False);

	return ldap_commit_mods("rid", rid_str, mods, False,
				lp_ldap_builtin_subcontext())
		&& ldap_incr_seq_num();
}


/*************************************************************************
  Return builtin aliases that a user is in.
 *************************************************************************/

static BOOL ldapbuiltin_getusergroups(const char *name,
			LOCAL_GRP **groups, int *num_grps)
{
	LOCAL_GRP *grouplist;
	fstring filter;
	int i;

	if(!ldap_connect())
		return (False);

	slprintf(filter, sizeof(pstring)-1,
#ifdef ENABLE_OLD_LDAP_SCHEMA
		 "(&(member=%s,*)(objectclass=sambaBuiltin))", name);
#else
		 "(&(sambaMember=%s,*)(objectclass=sambaBuiltin))", name);
#endif /* ENABLE_OLD_LDAP_SCHEMA */
	ldap_search_for(filter);

	*num_grps = i = ldap_count_entries(ldap_struct, ldap_results);

	if(!i) {
		*groups = NULL;
		ldap_disconnect();
		return (True);
	}

	*groups = grouplist = malloc(i * sizeof(LOCAL_GRP));
	do {
		i--;
	} while(ldapbuiltin_getgrp(&grouplist[i], NULL, NULL) && (i > 0));

	ldap_disconnect();
	return (True);
}


static struct aliasdb_ops ldapbuiltin_ops =
{
	ldapbuiltin_enumfirst,
	ldapbuiltin_enumclose,
	NULL,
	NULL,

	ldapbuiltin_getgrpbynam,
	NULL,
	ldapbuiltin_getgrpbyrid,
	ldapbuiltin_getcurrentgrp,

	ldapbuiltin_addgrp,
	ldapbuiltin_modgrp,
	ldapbuiltin_delgrp,

	ldapbuiltin_addmem,
	ldapbuiltin_delmem,

	ldapbuiltin_getusergroups
};

struct aliasdb_ops *ldap_initialise_builtin_db(void)
{
	/* 'ldap suffix' is required; if missing, try another db */
	char *p = lp_ldap_suffix();
	if (!p || !*p)
	{
		return NULL;
	}

	return &ldapbuiltin_ops;
}

#else
struct aliasdb_ops *ldap_initialise_builtin_db(void)
{
	return NULL;
}
#endif
