/* 
   Unix SMB/Netbios implementation.
   Version 2.0.
   LDAP domain group database for SAMBA
   Copyright (C) Matthew Chapman 1998
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
   
*/

#include "includes.h"
#include "libsamba.h"
#include "sids.h"
#include "samdb.h"
#include "others.h"

#ifdef USE_PASSDB_LDAP

#include <lber.h>
#include <ldap.h>


/* Internal state */
extern LDAP *ldap_struct;
extern LDAPMessage *ldap_results;
extern LDAPMessage *ldap_entry;

/* Static structure filled for requests */
static DOMAIN_GRP domgrp;


/***************************************************************
  Get group and membership information.
 ****************************************************************/

/* Returns the first value of the given attribute of an LDAP-entry . This is
   needed by the uniqueMember-stuff. */
static BOOL ldapgroup_getgrpmem_attr(LDAPMessage* entry, 
	const char* attribute, char* value) {
	char** values;
	int result = 0;

	values = ldap_get_values(ldap_struct, entry, attribute);

	if (values) {
		value[0] = '\0';
		
		result = convert_string(CH_UTF8, CH_UNIX, values[0], -1, 
			value, sizeof(pstring));

		ldap_value_free(values);
	}

	return (result > 0);
}

/* Resolves transparently (i.e. without impacting the global ldap_entry) a 
   user, adds it to the given DOMAIN_GRP_MEMBER-list and increments the 
   counter. This is needed by the uniqueMember-stuff. */
static void ldapgroup_resgrpmem(const char* dn, DOMAIN_GRP_MEMBER** members,
	int* num_membs) {
	
	LDAPMessage* results;
	LDAPMessage* entry;
	int err;

	fstring rid_str;
	fstring name;

	if (!dn)
		return;
	
	err = ldap_search_s(ldap_struct, dn, LDAP_SCOPE_BASE, 
		"(objectClass=sambaAccount)", NULL, 0, &results);
	
	if (err != LDAP_SUCCESS) {
		DEBUG(0, ("Error: search for user with dn %s failed.\n", dn));
		return;
	}

	entry = ldap_first_entry(ldap_struct, results);

	if (entry && 
		ldapgroup_getgrpmem_attr(entry, "rid", rid_str) &&
		(ldapgroup_getgrpmem_attr(entry, "ntuid", name) ||  
		 ldapgroup_getgrpmem_attr(entry, "uid", name))) {

		*members = realloc(*members, 
			(*num_membs + 1) * sizeof(DOMAIN_GRP_MEMBER));
		
		(*members)[*num_membs].rid  = strtol(rid_str, NULL, 16);
		fstrcpy((*members)[*num_membs].name, name);
		(*members)[*num_membs].attr = 0x7;

		(*num_membs)++;
	}
	else {
		DEBUG(0, ("Error: Couldn't resolve user with dn %s.\n", dn));
	}	

	if (results) 
		ldap_msgfree(results);
	
	return;
}


/* Compares two DOMAIN_GRP_MEMBER-structs by their RID and returns a value
   smaller 0 if a is smaller than b, 0 if they are equal and greater 0 if
   a is bigger than b. This is needed by the uniqueMember-stuff. */
static int ldapgroup_comparemember(const void* a, const void* b) {
	return (((const DOMAIN_GRP_MEMBER*)a)->rid - 
			((const DOMAIN_GRP_MEMBER*)b)->rid);
}


static DOMAIN_GRP *ldapgroup_getgrp(DOMAIN_GRP *group,
			     DOMAIN_GRP_MEMBER **members, int *num_membs)
{
	fstring temp;
	char **values;
	DOMAIN_GRP_MEMBER *memblist;
	DOMAIN_GRP_MEMBER* u_memblist; /* list of uniqueMembers */
	DOMAIN_GRP_MEMBER* m_memblist; /* merged result memblist and u_memblist */
	DOMAIN_GRP_MEMBER* next;       /* needed for the merging-algorithm */
	
	int u_num_membs;               /* number of uniqueMembers */
	int m_num_membs;               /* number of members in the merged array */
	char *value, *sep;
	int i, j;

	if(!ldap_entry)
		return NULL;

	if(!ldap_get_attribute("cn", group->name)) {
		DEBUG(0, ("Missing cn\n"));
                return NULL; }

	DEBUG(2,("Retrieving group [%s]\n", group->name));

        if(ldap_get_attribute("rid", temp)) {
		group->rid = strtol(temp, NULL, 16);
	} else {
		DEBUG(0, ("Missing rid\n"));
		return NULL;
	}

	if(!ldap_get_attribute("description", group->comment))
		group->comment[0] = 0;

	group->attr = 0x7;

	if(!members || !num_membs) {
		ldap_entry = ldap_next_entry(ldap_struct, ldap_entry);
		return group;
	}

#ifdef ENABLE_OLD_LDAP_SCHEMA
	if((values = ldap_get_values(ldap_struct, ldap_entry, "member"))) {
#else
	if((values = ldap_get_values(ldap_struct, ldap_entry, "sambaMember"))) {
#endif /* ENABLE_OLD_LDAP_SCHEMA */

		*num_membs = i = ldap_count_values(values);
		*members = memblist = malloc(i * sizeof(DOMAIN_GRP_MEMBER));

		do {
                        value = values[--i];
                
                        if(!(sep = strchr(value, ','))) {
                                DEBUG(0, ("Malformed group member\n"));
                                return NULL;
                        }
                        *(sep++) = 0;
                        fstrcpy(memblist[i].name, value);   

                        if(!(value = strchr(sep, ','))) {
                                DEBUG(0, ("Malformed group member\n"));
                                return NULL;
                        }
                        memblist[i].rid = strtol(sep, &value, 16);

			memblist[i].attr = 0x7;

		} while(i > 0);

		ldap_value_free(values);

	} else {
		*num_membs = 0;
		*members = NULL;
	}

	/* Get uniqueMembers from group-object: */
	if (lp_ldap_uniquemember() && 
		(values = ldap_get_values(ldap_struct, ldap_entry, "uniqueMember"))) {

		/* Following line isn't really needed but prevents gcc to complain
		   about uninitialized use of 'memblist': */
		memblist = *members;
		
		u_memblist = NULL;
		u_num_membs = 0;

		/* Resolve the DNs of the uniqheMember-entries: */
		for (i = ldap_count_values(values) - 1; i >= 0; i--) {
			ldapgroup_resgrpmem(values[i], &u_memblist, &u_num_membs);
		}

		ldap_value_free(values);

		/* Check whether both lists have members defined: */
		if (memblist && u_memblist) {
			/* Sort both arrays: */
			qsort(memblist  , *num_membs , sizeof(DOMAIN_GRP_MEMBER), 
				ldapgroup_comparemember);
			qsort(u_memblist, u_num_membs, sizeof(DOMAIN_GRP_MEMBER), 
				ldapgroup_comparemember);

			i = j = m_num_membs = 0;
			m_memblist = NULL;

			/* Merge the both arrays. A user can be present in one or both 
			   arrays. Take your time to check this algorithm... */
			for (m_num_membs = 0; 
				i < *num_membs || j < u_num_membs; 
				m_num_membs++) {
				
				/* Get the next entry for m_memblist: */
				if (i < *num_membs && j < u_num_membs) {
					if (memblist[i].rid == u_memblist[j].rid) {
						next = &memblist[i];
						i++;
						j++;
					}
					else if (memblist[i].rid < u_memblist[j].rid) {
						next = &memblist[i];
						i++;
					}
					else {
						next = &u_memblist[j];
						j++;
					}
				}
				else if (i < *num_membs) {
					next = &memblist[i];
					i++;
				}
				else {
					next = &u_memblist[j];
					j++;
				}

				m_memblist = realloc(m_memblist, 
					(m_num_membs + 1) * sizeof(DOMAIN_GRP_MEMBER));
				m_memblist[m_num_membs] = *next;
			}
			
			/* We don't need memblist and u_memblist anymore... :-) */
			free(memblist);
			free(u_memblist);
			*members = m_memblist;
			*num_membs = m_num_membs;
		}
		else if (u_memblist) {
			/* If we're here, memblist == NULL, so we don't have to free it */
			*members = u_memblist;
			*num_membs = u_num_membs;
		}
	}
	
	ldap_entry = ldap_next_entry(ldap_struct, ldap_entry);
	return group;
}


/************************************************************************
  Queues the necessary modifications to save a DOMAIN_GRP structure
 ************************************************************************/

static void ldapgroup_grpmods(DOMAIN_GRP *group, LDAPMod ***mods,
			      int operation)
{
	fstring temp;

	*mods = NULL;

	if(operation == LDAP_MOD_ADD) { /* immutable attributes */
		ldap_make_mod(mods, LDAP_MOD_ADD, "objectClass", "sambaGroup");
		ldap_make_mod(mods, LDAP_MOD_ADD, "cn", group->name);

		slprintf(temp, sizeof(temp)-1, "%x", group->rid);
		ldap_make_mod(mods, LDAP_MOD_ADD, "rid", temp);
	}

	ldap_make_mod(mods, operation, "description", group->comment);
}


/************************************************************************
  Create a group member entry
 ************************************************************************/

static BOOL ldapgroup_memmods(uint32 user_rid, LDAPMod ***mods, int operation)
{
	pstring member;
	fstring name;
	DOM_SID sid;
	uint32 type;
	char *dn = NULL;

	if (lp_ldap_uniquemember()
	    && ldap_connect())
	{
		if (ldap_search_by_rid(user_rid)
		    && ldap_entry)
		{
			dn = ldap_get_dn(ldap_struct, ldap_entry);
		}
		/* ldap_disconnect(); is done by the later functions anyway */
	}

	sid_copy(&sid, &global_sam_sid);
	sid_append_rid(&sid, user_rid);
	if (lookup_sid(&sid, name, &type))
		return (False);

	slprintf(member, sizeof(member)-1, "%s,%x,%d", name, user_rid, type);

	*mods = NULL;
#ifdef ENABLE_OLD_LDAP_SCHEMA
	ldap_make_mod(mods, operation, "member", member);
#else
	ldap_make_mod(mods, operation, "sambaMember", member);
#endif /* ENABLE_OLD_LDAP_SCHEMA */

	if (dn)
	{
		ldap_make_mod(mods, operation, "uniqueMember", dn);
		free(dn);
	}

	return True;
}


/***************************************************************
  Begin/end domain group enumeration.
 ****************************************************************/

static void *ldapgroup_enumfirst(BOOL update)
{
	int server_role = lp_server_role();

        if (server_role == ROLE_STANDALONE || server_role == ROLE_DOMAIN_MEMBER)
                return NULL;

	if (!ldap_connect())
		return NULL;

	ldap_search_for("objectclass=sambaGroup");

	return ldap_struct;
}

static void ldapgroup_enumclose(void *vp)
{
	ldap_disconnect();
}


/*************************************************************************
  Return information about domain groups and their members.
 *************************************************************************/

static DOMAIN_GRP *ldapgroup_getgrpbynam(const char *name,
	       DOMAIN_GRP_MEMBER **members, int *num_membs)
{
	fstring filter;
	DOMAIN_GRP *ret;

	if(!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(cn=%s)(objectClass=sambaGroup))", name);
	ldap_search_for(filter);

	ret = ldapgroup_getgrp(&domgrp, members, num_membs);

	ldap_disconnect();
	return ret;
}

static DOMAIN_GRP *ldapgroup_getgrpbyrid(uint32 grp_rid,
	       DOMAIN_GRP_MEMBER **members, int *num_membs)
{
	fstring filter;
	DOMAIN_GRP *ret;

	if(!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(rid=%x)(objectClass=sambaGroup))", grp_rid);
	ldap_search_for(filter);

	ret = ldapgroup_getgrp(&domgrp, members, num_membs);

	ldap_disconnect();
	return ret;
}

static DOMAIN_GRP *ldapgroup_getcurrentgrp(void *vp,
	       DOMAIN_GRP_MEMBER **members, int *num_membs)
{
	return ldapgroup_getgrp(&domgrp, members, num_membs);
}


/*************************************************************************
  Add/modify/delete domain groups.
 *************************************************************************/

static BOOL ldapgroup_addgrp(DOMAIN_GRP *group)
{
	LDAPMod **mods;

	if (!ldap_allocaterid(&group->rid))
	{
	    DEBUG(0,("RID generation failed\n"));
	    return (False);
	}

	ldapgroup_grpmods(group, &mods, LDAP_MOD_ADD); 
	return ldap_commit_mods("cn", group->name, mods, True,
				lp_ldap_groups_subcontext())
		&& ldap_incr_seq_num();
}

static BOOL ldapgroup_modgrp(DOMAIN_GRP *group)
{
	LDAPMod **mods;

	ldapgroup_grpmods(group, &mods, LDAP_MOD_REPLACE);
	return ldap_commit_mods("cn", group->name, mods, False,
				lp_ldap_groups_subcontext())
		&& ldap_incr_seq_num();
}

static BOOL ldapgroup_delgrp(uint32 grp_rid)
{
	fstring filter;
	char *dn;
	int err;

	if (!ldap_connect())
		return (False);

	slprintf(filter, sizeof(filter)-1,
		 "(&(rid=%x)(objectClass=sambaGroup))", grp_rid);
	ldap_search_for(filter);

	if (!ldap_entry || !(dn = ldap_get_dn(ldap_struct, ldap_entry)))
	{
		ldap_disconnect();
		return (False);
	}

	err = ldap_delete_s(ldap_struct, dn);
	free(dn);
	ldap_disconnect();

	if (err != LDAP_SUCCESS)
	{
		DEBUG(0, ("delete: %s\n", ldap_err2string(err)));
		return (False);
	}

	return True;
}


/*************************************************************************
  Add users to/remove users from groups.
 *************************************************************************/

static BOOL ldapgroup_addmem(uint32 grp_rid, uint32 user_rid)
{
	LDAPMod **mods;
        fstring rid_str;

	slprintf(rid_str, sizeof(rid_str)-1, "%x", grp_rid);

	if(!ldapgroup_memmods(user_rid, &mods, LDAP_MOD_ADD))
		return (False);

	return ldap_commit_mods("rid", rid_str, mods, False,
				lp_ldap_groups_subcontext())
		&& ldap_incr_seq_num();
}


/* This function removes deletions of attributes which aren't set so LDAP won't
   complain/fail on deletion. This is needed by the uniqueMember-stuff. */
static BOOL ldapgroup_checkdelmem(
	const char* grp_rid_str, LDAPMod*** modlist) {
	LDAPMod** mods;
	LDAPMod** new_mods;
	fstring filter;
	fstring value;
	int i, j;

	if (!ldap_connect())
		return (False);

	mods = *modlist;
	new_mods = NULL;
	
	/* Check mods for modifications which can't be done */
	for (i = 0; mods[i] != NULL; i++) {
		/* Get attribute-name of current modification: */
		if (!mods[i]->mod_type) 
			continue;
		
		/* For for every value if this value is set in the group-object: */
		for (j = 0; mods[i]->mod_values[j] != NULL; j++) {
			value[0] = '\0';
			
			/* We need to convert the value from UTF8 to ASCII back, because
			   ldap_make_mod will convert the value to UTF8... */
			if (convert_string(CH_UTF8, CH_UNIX, mods[i]->mod_values[j], -1,
				value, sizeof(fstring)) > 0) {
				
				/* If this search returns a result, we may delete the 
				   attribute */
				slprintf(filter, sizeof(filter) - 1,
					"(&(&(rid=%s)(objectClass=sambaGroup))(%s=%s))", 
					grp_rid_str, mods[i]->mod_type, value);
		
				if (ldap_search_for(filter) && ldap_entry) {
					/* The modification can be done, so will add to our 
					   new_mods-list: */
					ldap_make_mod(&new_mods, LDAP_MOD_DELETE, 
						mods[i]->mod_type, value);
				}		
			}
		}	
	}
	
	/* We don't need the old mods anymore, so we can free them and assign our
	   new mods. If there are no modifications to be done, it's OK too. That's
	   because ldap_commit_mod checks this and returns gracefully. */
	ldap_mods_free(*modlist, 1);
	*modlist = new_mods;

	/* Return false, if there are no modifications to do */
	return (new_mods != NULL);
}


static BOOL ldapgroup_delmem(uint32 grp_rid, uint32 user_rid)
{
	LDAPMod **mods;
        fstring rid_str;

	slprintf(rid_str, sizeof(rid_str)-1, "%x", grp_rid);

	if(!ldapgroup_memmods(user_rid, &mods, LDAP_MOD_DELETE))
		return (False);

	/* Delete modifications of non-existing attributes: */
	if (!ldapgroup_checkdelmem(rid_str, &mods))
		return (False);

	return ldap_commit_mods("rid", rid_str, mods, False,
				lp_ldap_groups_subcontext())
		&& ldap_incr_seq_num();
}


/*************************************************************************
  Return domain groups that a user is in.
 *************************************************************************/

static BOOL ldapgroup_getusergroups(const char *name, DOMAIN_GRP **groups,
				    int *num_grps)
{
	DOMAIN_GRP *grouplist;
	pstring filter;
	int i;
	char *dn = NULL;

	if(!ldap_connect())
		return (False);

	if (lp_ldap_uniquemember()
	    && samldap_search_user_by_ntname(name)
	    && ldap_entry)
	{
		dn = ldap_get_dn(ldap_struct, ldap_entry);
	}

	if (dn)
	{
		slprintf(filter, sizeof(filter)-1,
			 "(&(|(uniqueMember=%s)(sambaMember=%s,*))"
			   "(objectclass=sambaGroup))",
			 dn, name);
		free(dn);
	}
	else
	{
		slprintf(filter, sizeof(filter)-1,
#ifdef ENABLE_OLD_LDAP_SCHEMA
			 "(&(member=%s,*)(objectclass=sambaGroup))", name);
#else
			 "(&(sambaMember=%s,*)(objectclass=sambaGroup))", name);
#endif /* ENABLE_OLD_LDAP_SCHEMA */
	}

	ldap_search_for(filter);

	*num_grps = i = ldap_count_entries(ldap_struct, ldap_results);

	if(!i) {
		*groups = NULL;
		ldap_disconnect();
		return (True);
	}

	*groups = grouplist = malloc(i * sizeof(DOMAIN_GRP));
	do {
		i--;
	} while(ldapgroup_getgrp(&grouplist[i], NULL, NULL) && (i > 0));

	ldap_disconnect();
	return (True);
}


static struct groupdb_ops ldapgroup_ops =
{
	ldapgroup_enumfirst,
	ldapgroup_enumclose,
	NULL,
	NULL,

	ldapgroup_getgrpbynam,
	NULL,
	ldapgroup_getgrpbyrid,
	ldapgroup_getcurrentgrp,

	ldapgroup_addgrp,
	ldapgroup_modgrp,
	ldapgroup_delgrp,

	ldapgroup_addmem,
	ldapgroup_delmem,

	ldapgroup_getusergroups
};

struct groupdb_ops *ldap_initialise_group_db(void)
{
	/* 'ldap suffix' is required; if missing, try another db */
	char *p = lp_ldap_suffix();
	if (!p || !*p)
	{
		return NULL;
	}

	return &ldapgroup_ops;
}

#else
struct groupdb_ops *ldap_initialise_group_db(void)
{
	return NULL;
}
#endif

