/*
 * helper.c: functions for launching external helper scripts
 * Copyright 2003 by Peter Samuelson <peter@samba-tng.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/* MKPROTO: libsamba_proto.h		(do not remove this line) */

#include "includes.h"
#include <stdarg.h>
#include "libsamba.h"
#include "tng_misc.h"

/* #define HELPER_LOG_STDERR */

/*********************************************************************
 * These functions set up and execute child processes in the 'script
 * directory', normally '{confdir}/scripts/'.  For efficiency and
 * general sanity, we define a few constraints on these scripts:
 *
 * (a) We don't make any attempt at a "safe" environment.  Effective
 *     UID/GID are defined by the caller - may be root, may be the
 *     connected user, may be nobody.  No chroot, no sanitizing of
 *     environment vars.
 *
 * (b) The script must ignore extra open file descriptors other than
 *     std{in,out,err}.  It is not efficient to painstakingly close all
 *     fds before executing the child.  Script writers should be aware
 *     of the information leakage / DoS potential.
 *
 * (c) We don't accept more than 8 command-line args.  This is easy to
 *     relax, should there be a need.
 *********************************************************************/

struct child {
	struct child *next;
	char *name;
	pid_t pid;
	int cookie;
	int stdinfd, stdoutfd;
};
static struct child *child;
static int child_last_cookie;


/*
 * move half a pipe to 'fd', and close the other half.
 * 'which' is the half you want to keep (0 or 1).
 */
static void dup_half_pipe(int p[], int which, int fd)
{
	close(p[1-which]);
	p[1-which] = -1;
	if (p[which] != fd) {
		close(fd);
		dup2(p[which], fd);
		close(p[which]);
		p[which] = -1;
	}
}

/*
 * open /dev/null as a specific fd
 */
static void open_devnull(int fd, int flags)
{
	int newfd;

	close(fd);
	newfd = sys_open("/dev/null", flags, 0);
	if (newfd < 0) {
		DEBUG(0, ("Can't open /dev/null\n"));
		return;
	}
	if (newfd != fd) {
		dup2(newfd, fd);
		close(fd);
	}
}

static void close_pipe(int p[])
{
	int i;
	for (i=0; i<2; i++)
		if (p[i] >= 0) {
			close(p[i]);
			p[i] = -1;
		}
}


/*
 * Opens a helper script, returns stdin and stdout by reference.
 * stderr is logged via DEBUG(2), from another child process.
 *
 * NULL means the given fd should be closed instead.
 * Returns cookie > 0 for success, 0 for script not found, -1 for other problem.
 */
int helper_spawnv(int *stdinfd, int *stdoutfd, char *argv[])
{
	/* pipes are named from child's perspective */
	int inpipe[] = {-1,-1}, outpipe[] = {-1,-1};
#ifdef HELPER_LOG_STDERR
	int errpipe[] = {-1,-1};
#endif
	fstring exe;
	SMB_STRUCT_STAT st;
	struct child *new_child;

	fstrcpy(exe, lp_scriptdir());
	fstrcat(exe, "/");
	fstrcat(exe, argv[0]);

	/* don't bother to fork/exec if the script doesn't exist */
	if (sys_stat(exe, &st) < 0 ||
	    !S_ISREG(st.st_mode) ||
	    !(st.st_mode & 0111))
		return 0;

	new_child = g_new(struct child, 1);
	if (!new_child) {
		DEBUG(0,("helper_spawnv: out of memory\n"));
		return -1;
	}
	new_child->name = argv[0];
	if (++child_last_cookie < 1)
		child_last_cookie = 1;
	new_child->cookie = child_last_cookie;

	if (stdinfd && pipe(inpipe)) {
		DEBUG(0,("helper_spawnv: stdin pipe failed\n"));
		goto err0;
	}

	if (stdoutfd && pipe(outpipe)) {
		DEBUG(0,("helper_spawnv: stdout pipe failed\n"));
		goto err1;
	}

	CatchChildLeaveStatus();
	new_child->pid = sys_fork();
	if (new_child->pid < 0) {
		DEBUG(0,("helper_spawnv: fork failed\n"));

		close_pipe(outpipe);
	err1:
		close_pipe(inpipe);
	err0:
		free(new_child);
		return -1;
	}

	if (new_child->pid > 0) {
		/* we are the parent */
		new_child->next = child;
		new_child->stdinfd = inpipe[1];
		new_child->stdoutfd = outpipe[0];
		if (stdinfd)
			*stdinfd = new_child->stdinfd;
		if (stdoutfd)
			*stdoutfd = new_child->stdoutfd;
		close(inpipe[0]);
		close(outpipe[1]);
		child = new_child;
		return child->cookie;
	}

	/* we are the child */

#ifdef HELPER_LOG_STDERR
	if (pipe(errpipe)) {
		DEBUG(2,("helper_spawnv: stderr pipe failed\n"));
		open_devnull(STDERR_FILENO, O_WRONLY);
	}
	else {
		pid_t errp = sys_fork();
		if (errp < 0)
			DEBUG(2,("helper_spawnv: second fork failed\n"));
		else if (errp == 0) {

			/* we are the stderr daemon */
			int start = 0, end = 0;
			pstring buf;

			close_pipe(inpipe);
			close_pipe(outpipe);

			close(errpipe[1]);

			/* basically, DEBUG('s/^/spawn argv[0]: /') */
			while (1) {
				int len = read(errpipe[0], buf+end, sizeof(buf)-end-2);
				char *p;

				if (len <= 0) {
					if (end > start)
						DEBUG(3,("spawn %s: %s\n", argv[0], buf+start));
					exit(0);
				}

				end += len;
				for (p = strchr(buf+start, '\n'); p; p = strchr(p, '\n')) {
					*p++ = '\0';
					DEBUG(3,("spawn %s: %s\n", argv[0], buf+start));
					start = p-buf;
				}
				if (end > start)
					memcpy(buf, buf+start, end-start);
				end -= start;
				start = 0;
				buf[end] = '\0';
			}
		}

		/* back to the child */
		dup_half_pipe(errpipe, 1, STDERR_FILENO);
	}
#else
	open_devnull(STDERR_FILENO, O_WRONLY);
#endif
	if (stdinfd)
		dup_half_pipe(inpipe, 0, STDIN_FILENO);
	else
		open_devnull(STDIN_FILENO, O_RDONLY);

	if (stdoutfd)
		dup_half_pipe(outpipe, 1, STDOUT_FILENO);
	else
		open_devnull(STDOUT_FILENO, O_WRONLY);

	execv(exe, argv);
	DEBUG(0,("helper_spawnv: exec '%s' failed\n", exe));
	exit(1);
}

/* varargs version of the above. */
int helper_spawnl(int *stdinfd, int *stdoutfd, char *cmd, ...)
{
	char *argv[10];
	int i;
	va_list ap;

	argv[0] = cmd;
	argv[9] = NULL;

	va_start(ap, cmd);
	for (i=1; i<10; i++) {
		argv[i] = va_arg(ap, char *);
		if (!argv[i])
			break;
	}
	va_end(ap);

	return helper_spawnv(stdinfd, stdoutfd, argv);
}

/*
 * wait for child process
 */
int helper_sync(int cookie)
{
	struct child *c, *c_old = NULL;
	pid_t pid;
	int status;

	if (cookie <= 0)
		return cookie;

	for (c = child; c; c = c->next) {
		if (cookie == c->cookie)
			break;
		c_old = c;
	}
	if (!c) {
		DEBUG(0,("helper_sync: unknown cookie %d\n", cookie));
		return -1;
	}

	if (c->stdinfd >= 0)
		close(c->stdinfd);
	if (c->stdoutfd >= 0)
		close(c->stdoutfd);
	pid = sys_waitpid(c->pid, &status, 0);
	if (pid != c->pid) {
		DEBUG(0,("helper_sync: error syncing '%s'\n", c->name));
		status = -1;
	}
	if (c_old)
		c_old->next = c->next;
	else
		child = c->next;
	free(c);

	if (!child)
		/* reset the signal handler, which was changed by _spawn() */
		CatchChild();

	return status;
}

/*
 * Spawns a helper process, feeds its stdin from a buffer,
 * and collects its stdout into a static buffer.
 * Note that the parameters 'inbuf' and 'outbuf' are from the
 * perspective of the child process.
 */
int helper_spawnv_sync(const char *inbuf, size_t inbuflen, char **outbuf, size_t *outbuflen, char *argv[])
{
	int fdin, fdout;
	int cookie;
	static pstring out_buf;
	size_t in_len = 0, out_len = 0;

	if (!inbuf)
		fdin = -1;
	if (!outbuf)
		fdout = -1;

	cookie = helper_spawnv(inbuf  ? &fdin  : NULL,
			       outbuf ? &fdout : NULL, argv);
	if (cookie <= 0)
		return cookie;

	while (fdin > -1 && fdout > -1) {
		fd_set fds;
		int fdmax = fdin > fdout ? fdin : fdout;

		FD_ZERO(&fds);
		if (fdin > -1)
			FD_SET(fdin, &fds);

		if (fdout > -1)
			FD_SET(fdout, &fds);

		sys_select_intr(fdmax, &fds, NULL);

		if (FD_ISSET(fdin, &fds)) {
			size_t len = inbuflen - in_len;
			len = write(fdin, inbuf+in_len, len);
			if (len > 0) {
				in_len += len;
				if (in_len == inbuflen)
					fdin = -1;
			}
			else
				fdin = -1;
		}
		if (FD_ISSET(fdout, &fds)) {
			size_t len = sizeof(out_buf) - out_len;
			len = read(fdout, out_buf+out_len, len);
			if (len > 0) {
				out_len += len;
				if (out_len == sizeof(out_buf)) {
					DEBUG(2,("helper_spawnv_sync: truncating output\n"));
					fdout = -1;
				}
			}
			else
				fdout = -1;
		}
	}
	if (outbuf)
		*outbuf = out_buf;
	if (outbuflen)
		*outbuflen = out_len;

	return helper_sync(cookie);
}

int helper_spawnl_sync(char *inbuf, int inbuflen, char **outbuf, int *outbuflen, char *cmd, ...)
{
	char *argv[10];
	int i;
	va_list ap;

	argv[0] = cmd;
	argv[9] = NULL;

	va_start(ap, cmd);
	for (i=1; i<10; i++) {
		argv[i] = va_arg(ap, char *);
		if (!argv[i])
			break;
	}
	va_end(ap);

	return helper_spawnv_sync(inbuf, inbuflen, outbuf, outbuflen, argv);
}
