/* 
   Unix SMB/Netbios implementation.
   Version 1.9.
   Samba utility functions
   Copyright (C) Andrew Tridgell 1992-2000
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
/* MKPROTO: lib_smb_proto.h		(do not remove this line) */

#include "includes.h"
#include "byteorder.h"
#include "libsamba.h"
#include "libsmb.h"

#undef DBGC_CLASS
#define DBGC_CLASS DBGC_SMB

extern int smb_read_error;


/*******************************************************************
  show a smb message structure
********************************************************************/
void show_msg(char *buf)
{
	int i;
	int bcc = 0;

	if (! DEBUGLVL(5))
		return;

	DEBUGADD(5,
	      ("size=%d\nsmb_com=0x%x\nsmb_rcls=%d\nsmb_reh=%d\nsmb_err=%d\nsmb_flg=%d\nsmb_flg2=%d\n",
	       smb_len(buf),
	       (int)CVAL(buf, smb_com), (int)CVAL(buf, smb_rcls),
	       (int)CVAL(buf, smb_reh), (int)SVAL(buf, smb_err),
	       (int)CVAL(buf, smb_flg), (int)SVAL(buf, smb_flg2)));
	DEBUGADD(5,
	      ("smb_tid=%d\nsmb_pid=%d\nsmb_uid=%d\nsmb_mid=%d\nsmt_wct=%d\n",
	       (int)SVAL(buf, smb_tid), (int)SVAL(buf, smb_pid),
	       (int)SVAL(buf, smb_uid), (int)SVAL(buf, smb_mid),
	       (int)CVAL(buf, smb_wct)));

	for (i = 0; i < (int)CVAL(buf, smb_wct); i++)
	{
		DEBUGADD(5, ("smb_vwv[%d]=%d (0x%X)\n", i,
			  SVAL(buf, smb_vwv + 2 * i), SVAL(buf,
							   smb_vwv + 2 * i)));
	}

	bcc = (int)SVAL(buf, smb_vwv + 2 * (CVAL(buf, smb_wct)));

	DEBUGADD(5, ("smb_bcc=%d\n", bcc));

	if (bcc == 0)
		return;

	if (! DEBUGLVL(10))
		return;

	if (! DEBUGLVL(50))
	{
		bcc = MIN(bcc, 512);
	}

	dump_data(10, smb_buf(buf), bcc);
}

/****************************************************************************
  send an smb to a fd 
****************************************************************************/

BOOL send_smb(int fd, char *buffer)
{
	size_t len;
	size_t nwritten = 0;
	ssize_t ret;
	len = smb_len(buffer) + 4;

	while (nwritten < len)
	{
		ret = write_socket(fd, buffer + nwritten, len - nwritten);
		if (ret <= 0)
		{
			DEBUG(0,
			      ("Error writing %d bytes to client. %d. Exiting\n",
			       (int)len, (int)ret));
			exit(1);
		}
		nwritten += ret;
	}

	return True;
}

/****************************************************************************
read 4 bytes of a smb packet and return the smb length of the packet
store the result in the buffer
This version of the function will return a length of zero on receiving
a keepalive packet.
timeout is in milliseconds.
****************************************************************************/

static ssize_t read_smb_length_return_keepalive(int fd, char *inbuf,
						unsigned int timeout)
{
	ssize_t len = 0;
	int msg_type;
	BOOL ok = False;

	while (!ok)
	{
		if (timeout > 0)
			ok = (read_socket_with_timeout(fd, inbuf, 4, 4, timeout) == 4);
		else
			ok = (read_socket_data(fd, inbuf, 4) == 4);

		if (!ok)
			return (-1);

		len = smb_len(inbuf);
		msg_type = CVAL(inbuf, 0);

		if (msg_type == 0x85)
			DEBUG(5, ("Got keepalive packet\n"));
	}

	DEBUG(10, ("got smb length of %d\n", len));

	return (len);
}

/****************************************************************************
read 4 bytes of a smb packet and return the smb length of the packet
store the result in the buffer. This version of the function will
never return a session keepalive (length of zero).
timeout is in milliseconds.
****************************************************************************/
ssize_t read_smb_length(int fd, char *inbuf, unsigned int timeout)
{
	ssize_t len;

	for (;;)
	{
		len = read_smb_length_return_keepalive(fd, inbuf, timeout);

		if (len < 0)
			return len;

		/* Ignore session keepalives. */
		if (CVAL(inbuf, 0) != 0x85)
			break;
	}

	DEBUG(10, ("read_smb_length: got smb length of %d\n", len));

	return len;
}

/****************************************************************************
  read an smb from a fd. Note that the buffer *MUST* be of size
  BUFFER_SIZE+SAFETY_MARGIN.
  The timeout is in milliseconds. 
  This function will return on a
  receipt of a session keepalive packet.
****************************************************************************/
BOOL receive_smb(int fd, char *buffer, unsigned int timeout)
{
	ssize_t len, ret;

	smb_read_error = 0;

	memset(buffer, '\0', smb_size + 100);

	len = read_smb_length_return_keepalive(fd, buffer, timeout);
	if (len < 0)
	{
		DEBUG(10, ("receive_smb: length < 0!\n"));
		return (False);
	}

	if (len > BUFFER_SIZE)
	{
		DEBUG(0, ("Invalid packet length! (%d bytes).\n", len));
		if (len > BUFFER_SIZE + (SAFETY_MARGIN / 2))
		{
			exit(1);
		}
	}

	if (len > 0)
	{
		ret = read_socket_data(fd, buffer + 4, len);
		if (ret != len)
		{
			smb_read_error = READ_ERROR;
			return False;
		}
	}

	return (True);
}

