/* 
 * regdump.c: registry dump / restore commands for Samba-TNG
 * Copyright (C) 2002, 2003 by Peter Samuelson <peter@samba-tng.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "includes.h"
#include <errno.h>
#include "libsamba.h"
#include "rpcclient.h"
#include "rpc_client.h"
#include "rpc_reg.h"
#include "nterr.h"
#include "libsmb.h"

extern FILE *out_hnd;

static FILE *dump_fout;
static int dump_header_printed;
static fstring dump_srvname;

static void dump_recurse(POLICY_HND *, const char *);

static char *dump_strquote(const char *str)
{
	static pstring ret;
	int i, j;
	for (i=j=0; j<sizeof(ret)-2; i++,j++)
	{
		if(str[i] == '\\' || str[i] == '"')
		{
			ret[j++] = '\\';
		}
		ret[j] = str[i];
	}
	ret[j] = '\0';
	return ret;
}

static NTSTATUS nt4_reg_headers(POLICY_HND *ignored, int sw, const char *keyname, int num)
{
	if (!dump_header_printed)
	{
		report(dump_fout, "REGEDIT4\r\n");
		dump_header_printed = 1;
	}
	if (sw == 2)
		report(dump_fout, "\r\n[%s]\r\n", keyname);
	return NT_STATUS_OK;
}
static NTSTATUS nt4_reg_key(POLICY_HND *handle,
			    const char *full_name, const char *name, time_t modtime)
{
	pstring real_name;
	POLICY_HND subhnd;

	if (!reg_open_entry(handle, name,
			    SEC_RIGHTS_MAXIMUM_ALLOWED, &subhnd))
		return NT_STATUS_UNSUCCESSFUL;

	safe_strconcat(real_name, sizeof(pstring), full_name, "\\", name, NULL);
	dump_recurse(&subhnd, real_name);

	reg_close(&subhnd);
	return NT_STATUS_OK;
}
static NTSTATUS nt4_reg_val(POLICY_HND *handle,
			    const char *full_name, const char *name,
			    uint32 type, const BUFFER2 *const value)
{
	char *typename;
	int linelen, i;
	int buflen = value->buf_len;

	if (strlen(name))
	{
		char *valname = dump_strquote(name);
		linelen = strlen(valname) + 3;
		report(dump_fout, "\"%s\"=", valname);
	}
	else
	{
		report(dump_fout, "@=");
		linelen = 2;
	}
	switch(type)
	{
		case REGVAL_NONE:
		{
			return NT_STATUS_OK;
		}
		case REGVAL_SZ:
		{
			pstring valstr;
			unibuf_to_ascii(valstr, value->buffer,
					MIN(buflen, sizeof(valstr)-1));
			report(dump_fout, "\"%s\"\r\n", dump_strquote(valstr));
			return NT_STATUS_OK;
		}
		case REGVAL_DWORD:
		{
			report(dump_fout, "dword:%08x\r\n", buffer2_to_uint32(value));
			return NT_STATUS_OK;
		}
		case REGVAL_BINARY:
		{
			typename = "hex";
			break;
		}
		default:
		{
			static char defaulttype[16];
			slprintf(defaulttype, 15, "hex(%x)", type);
			typename = defaulttype;
			break;
		}
	}
	report(dump_fout, "%s:", typename);
	linelen += strlen(typename) + 1;
	for (i=0; i<buflen; i++)
	{
		int islastbyte = (i==buflen-1);
		if (linelen > 75 + islastbyte)
		{
			report(dump_fout, "\\\r\n  ");
			linelen = 2;
		}
		report(dump_fout, "%02x%s",
		       value->buffer[i], islastbyte ? "\r\n" : ",");
		linelen += 2;
	}
	return NT_STATUS_OK;
}



static struct {
	char *name;
	BOOL valsfirst;
	RegKeyFn reg_fn;
	RegSubKeyFn reg_key_fn;
	RegSubValueFn reg_val_fn;
} dump_formats[] = {
	/* more formats to come in the future - at least nt5 */
	{ "nt4", True, nt4_reg_headers, nt4_reg_key, nt4_reg_val },
};
static int fmt_idx;

static void dump_recurse(POLICY_HND *handle, const char *key)
{
	msrpc_reg_enum_key(handle, key,
			   dump_formats[fmt_idx].valsfirst,
			   dump_formats[fmt_idx].reg_fn,
			   dump_formats[fmt_idx].reg_key_fn,
			   dump_formats[fmt_idx].reg_val_fn);
}

static void recursive_dump(const char *key)
{
	msrpc_reg_enum_keyname(dump_srvname, key,
			       dump_formats[fmt_idx].valsfirst,
			       dump_formats[fmt_idx].reg_fn,
			       dump_formats[fmt_idx].reg_key_fn,
			       dump_formats[fmt_idx].reg_val_fn);
}

NTSTATUS cmd_reg_dump (struct client_info *info, int argc, char *argv[])
{
	char *format = "nt4";
	char *name = argv[0];

	dump_fout = out_hnd;
	dump_header_printed = 0;
	slprintf(dump_srvname, sizeof(dump_srvname)-1, "\\\\%s", info->dest_host);
	strupper(dump_srvname);

	argv++; argc--;
	for (; argc && *argv[0] == '-'; argv++,argc--)
	{
		if (!strcmp(argv[0], "-o"))
		{
			dump_fout = fopen(argv[1], "w");
			if (!dump_fout)
			{
				report(out_hnd, "%s: cannot open input file `%s': %s\n",
				       name, argv[1], strerror(errno));
				return map_nt_error_from_unix(errno);
			}
			argv++; argc--;
		}
		else if (!strcmp(argv[0], "-t"))
		{
			format = argv[1];
			argv++; argc--;
		}
		else
		{
			cmd_help_current(info);
			return NT_STATUS_INVALID_PARAMETER;
		}
	}
	if (!argc)
	{
		cmd_help_current(info);
		return NT_STATUS_OK;
	}
	for (fmt_idx = 0;; fmt_idx++)
	{
		if (!strcmp(dump_formats[fmt_idx].name, format))
		{
			break;
		}
		if (!dump_formats[fmt_idx].name)
		{
			report(out_hnd, "%s: invalid format %s\n",
			       name, format);
			return NT_STATUS_INVALID_PARAMETER;
		}
	}

	recursive_dump(argv[0]);

	if (dump_fout != out_hnd)
	{
		fclose(dump_fout);
	}
	return NT_STATUS_OK;
}

NTSTATUS cmd_reg_restore (struct client_info *info, int argc, char *argv[])
{
	char *format = NULL;

	argv++; argc--;
	for (; argc && *argv[0] == '-'; argv++,argc--)
	{
		if (!strcmp(argv[0], "-t"))
		{
			format = argv[1];
			argv++; argc--;
		}
		else
		{
			cmd_help_current(info);
			return NT_STATUS_INVALID_PARAMETER;
		}
	}
	if (argc != 2)
	{
		cmd_help_current(info);
		return NT_STATUS_OK;
	}
	report(out_hnd, "reg_restore not yet functional\n");
	return NT_STATUS_NOT_IMPLEMENTED;
}
